from rest_framework import serializers
from rest_framework.exceptions import ValidationError

from garpix_notify.models import SystemNotify
from garpix_notify.models.choices import TYPE, STATE


class SystemNotifySerializer(serializers.ModelSerializer):
    class Meta:
        model = SystemNotify
        fields = '__all__'


class ReadSystemNotifySerializer(serializers.Serializer):
    ids = serializers.ListField(child=serializers.IntegerField(), allow_empty=False)

    def validate_ids(self, value):
        db_ids = SystemNotify.objects.filter(user=self.context['request'].user, type=TYPE.SYSTEM, state=STATE.DELIVERED,
                                             is_read=False).values_list('id', flat=True)

        for id_val in value:
            if id_val not in db_ids:
                raise ValidationError(f"Уведомление с id {id_val} не муществует")

        return value



from fcm_django.models import FCMDevice


class NotifyDevice(FCMDevice):
    """
    Переопределен Класс для того чтобы спрятать модель в другое приложение
    """

    class Meta:
        proxy = True
        verbose_name = 'FCM аккаунт'
        verbose_name_plural = 'FCM аккаунты'



from django.db import models
from solo.models import SingletonModel

from garpix_notify.models.choices import EMAIL_MALLING, PARSE_MODE_TELEGRAM, SMS_URL, CALL_URL


class NotifyConfig(SingletonModel):
    periodic = models.IntegerField(default=60, verbose_name='Периодичность отправки уведомлений (сек.)')

    email_max_day_limit = models.IntegerField(default=240, verbose_name='Дневной лимит отправки писем')
    email_max_hour_limit = models.IntegerField(default=40, verbose_name='Часовой лимит отправки писем')

    sms_url_type = models.IntegerField(default=SMS_URL.SMSRU_ID, choices=SMS_URL.TYPES,
                                       verbose_name='URL СМС провайдера')
    sms_api_id = models.CharField(default='1234567890', blank=True, max_length=255,
                                  verbose_name='API ID СМС провайдера')
    sms_login = models.CharField(default='', blank=True, max_length=255,
                                 verbose_name='Логин пользователя СМС провайдера')
    sms_password = models.CharField(default='', blank=True, max_length=255,
                                    verbose_name='Пароль для api СМС провайдера')
    sms_from = models.CharField(default='', blank=True, max_length=255, verbose_name='Отправитель СМС',
                                help_text='Например, Garpix')
    call_url_type = models.IntegerField(default=CALL_URL.SMSRU_CALL_API_ID, choices=CALL_URL.TYPES,
                                        verbose_name='URL звонка провайдера')
    call_api_id = models.CharField(default='1234567890', blank=True, max_length=255,
                                   verbose_name='API ID оператора связи')
    call_login = models.CharField(default='', blank=True, max_length=255,
                                  verbose_name='Логин/Индетификатор сервиса оператора связи')
    call_password = models.CharField(default='', blank=True, max_length=255,
                                     verbose_name='Пароль/Секретный ключ оператора связи')
    telegram_api_key = models.CharField(default='000000000:AAAAAAAAAA-AAAAAAAA-_AAAAAAAAAAAAAA', blank=True,
                                        max_length=255, verbose_name='Telegram API Key')
    telegram_bot_name = models.CharField(default='', blank=True, help_text='Например, MySuperBot',
                                         max_length=255, verbose_name='Telegram Имя бота')
    telegram_welcome_text = models.TextField(blank=True,
                                             default='Добрый день! Здесь вы можете получать уведомления от нашего сайта',
                                             verbose_name='Telegram - Приветственный текст бота')
    telegram_help_text = models.TextField(blank=True,
                                          default='Используйте команду /set <уникальный код> для того, чтобы получать сообщения от бота. Уникальный код вы можете получить на нашем сайте',
                                          verbose_name='Telegram - Текст помощи бота')
    telegram_bad_command_text = models.TextField(blank=True,
                                                 default='Неправильный формат команды',
                                                 verbose_name='Telegram - Текст неправильной команды бота')
    telegram_success_added_text = models.TextField(blank=True,
                                                   default='Ваша учетная запись успешно привязана к боту. Вы будете получать уведомления!',
                                                   verbose_name='Telegram - Текст успешно добавлен код')
    telegram_failed_added_text = models.TextField(blank=True,
                                                  default='Ошибка при привязке учетной записи. Пожалуйста, свяжитесь с техподдержкой',
                                                  verbose_name='Telegram - Текст провал, не добавлен код')
    telegram_parse_mode = models.CharField(default=PARSE_MODE_TELEGRAM.EMPTY, choices=PARSE_MODE_TELEGRAM.TYPES,
                                           verbose_name='Тип парсера телеграм сообщений', max_length=100, blank=True)
    telegram_disable_notification = models.BooleanField(verbose_name='Пользователи получат уведомление без звука',
                                                        default=False)
    telegram_disable_web_page_preview = models.BooleanField(
        verbose_name='Отключает предварительный просмотр ссылок в сообщениях',
        default=False)
    telegram_allow_sending_without_reply = models.BooleanField(
        verbose_name='Разрешить, если сообщение должно быть отправлено, даже если ответное сообщение не найдено',
        default=False)
    telegram_timeout = models.FloatField(default=None, blank=True, verbose_name="Тайм-аут чтения с сервера", null=True)
    viber_api_key = models.CharField(default='000000000:AAAAAAAAAA-AAAAAAAA-_AAAAAAAAAAAAAA', blank=True,
                                     max_length=255, verbose_name='Viber API Key')
    viber_bot_name = models.CharField(blank=True, max_length=255, verbose_name='Название viber бота',
                                      default='Viber bot')
    whatsapp_sender = models.CharField(max_length=30, blank=True, default='',
                                       verbose_name='Телефон отправителя WhatsApp')
    whatsapp_auth_token = models.CharField(default='XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX', blank=True,
                                           max_length=255, verbose_name='WhatsApp Auth Token')
    whatsapp_account_sid = models.CharField(default='ACXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX', blank=True,
                                            max_length=255, verbose_name='WhatsApp Account SID')
    is_email_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку Email')
    is_sms_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку SMS')
    is_call_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку звонков')
    is_push_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку PUSH')
    is_telegram_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку Telegram')
    is_viber_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку Viber')
    is_whatsapp_enabled = models.BooleanField(default=True, verbose_name='Разрешить отправку WhatsApp')

    viber_success_added_text = models.TextField(blank=True,
                                                default='Ваша учетная запись успешно привязана к боту. Вы будете получать уведомления!',
                                                verbose_name='Viber - Текст успешно добавлен код')
    viber_failed_added_text = models.TextField(blank=True,
                                               default='Ошибка при привязке учетной записи. Пожалуйста, свяжитесь с техподдержкой',
                                               verbose_name='Viber - Текст провал, не добавлен код')

    viber_text_for_new_sub = models.TextField(blank=True,
                                              default='cпасибо за подписку, Введите secret_key чтобы получать сообщения от бота.',
                                              verbose_name='Viber - Текст  для новых подписчиков')

    viber_welcome_text = models.TextField(blank=True,
                                          default='для активации бота нужно отправить любое сообщения',
                                          verbose_name='Viber - Приветственный текст бота')
    email_malling = models.IntegerField(default=EMAIL_MALLING.BCC, choices=EMAIL_MALLING.TYPES,
                                        verbose_name='Тип массовой рассылки',
                                        help_text='Если выбрана обычная рассылка, то пользователи будут видеть email друг друга')

    class Meta:
        verbose_name = 'Настройка'
        verbose_name_plural = 'Настройки'

    def __str__(self):
        return 'Настройки'



import json
import os
import re
from datetime import datetime
from typing import Optional, List

import requests
from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import models, transaction
from django.db.models import Manager
from django.utils.html import format_html
from django.utils.module_loading import import_string
from django.db.models.signals import post_save
from django.dispatch import receiver

from .log import NotifyErrorLog
from .user_list import NotifyUserList
from .category import NotifyCategory
from .choices import TYPE, STATE, StatusMessage
from .file import NotifyFile
from .template import NotifyTemplate
from ..exceptions import IsInstanceException
from ..mixins import UserNotifyMixin
from ..utils.send_data import SendData

from ..clients import SMSClient, EmailClient, CallClient, TelegramClient, ViberClient, PushClient, WhatsAppClient

NotifyMixin = import_string(getattr(settings, 'GARPIX_NOTIFY_MIXIN', 'garpix_notify.mixins.notify_mixin.NotifyMixin'))

User = get_user_model()


class Notify(NotifyMixin, UserNotifyMixin):
    """
    Уведомление
    """
    subject = models.CharField(max_length=255, default='', blank=True, verbose_name='Тема')
    text = models.TextField(verbose_name='Текст')
    html = models.TextField(verbose_name='HTML', blank=True, default='')

    user: User = models.ForeignKey(settings.AUTH_USER_MODEL, blank=True, null=True, on_delete=models.SET_NULL,
                                   related_name='notifies', verbose_name='Пользователь (получатель)')
    email = models.EmailField(max_length=255, blank=True, null=True, verbose_name='Email Получателя',
                              help_text='Используется только в случае отсутствия указанного пользователя')
    sender_email = models.EmailField(max_length=255, blank=True, null=True, verbose_name='Email Отправителя')

    state = models.IntegerField('Состояние', choices=STATE.CHOICES, default=STATE.WAIT)
    event = models.IntegerField('Событие', choices=settings.CHOICES_NOTIFY_EVENT, blank=True, null=True)
    room_name = models.CharField('Название комнаты', max_length=255, null=True, blank=True)
    type = models.IntegerField(choices=TYPE.CHOICES, verbose_name='Тип')
    category = models.ForeignKey(NotifyCategory, on_delete=models.CASCADE, related_name='notifies',
                                 verbose_name='Категория')

    files = models.ManyToManyField(NotifyFile, verbose_name='Файлы')

    is_read = models.BooleanField(default=False, verbose_name='Прочитано')
    data_json = models.TextField(blank=True, null=True, verbose_name='Данные пуш-уведомления (JSON)')

    users_list = models.ManyToManyField(NotifyUserList, blank=True, verbose_name='Списки пользователей для рассылки')

    send_at = models.DateTimeField(blank=True, null=True, verbose_name='Время начала отправки')
    created_at = models.DateTimeField('Дата создания', auto_now_add=True)
    sent_at = models.DateTimeField('Дата отправки', blank=True, null=True)

    is_delete_after = models.BooleanField(default=False, verbose_name='Удалять после отправки')

    objects = Manager()

    def __str__(self):
        return self.subject if self.subject and self.subject != '' else f'Уведомление № {self.id}'

    def _get_sender(self):
        if self.user:
            self.email = self.user.email if self.user.email else self.email
            self.phone = str(self.user.phone) if self.user.phone else str(self.phone)
            self.telegram_chat_id = self.user.telegram_chat_id if self.user.telegram_chat_id else self.telegram_chat_id
            self.viber_chat_id = self.user.viber_chat_id if self.user.viber_chat_id else self.viber_chat_id

        if self.phone is not None:
            self.phone = re.sub(r"\D", "", self.phone)

    def start_send(self):  # noqa

        # Если передан пользователь, то перезаписываем данные (если они есть у пользователя)
        self._get_sender()
        if self.type == TYPE.EMAIL:
            EmailClient.send_email(self)
        elif self.type == TYPE.SMS:
            SMSClient.send_sms(self)
        elif self.type == TYPE.PUSH:
            PushClient.send_push(self)
        elif self.type == TYPE.TELEGRAM:
            TelegramClient.send_telegram(self)
        elif self.type == TYPE.VIBER:
            ViberClient.send_viber(self)
        elif self.type == TYPE.CALL:
            CallClient.send_call(self)
        elif self.type == TYPE.WHATSAPP:
            WhatsAppClient.send_whatsapp(self)
        if self.is_delete_after and self.state == STATE.DELIVERED:
            self._delete_notify()
        else:
            self.save()

    @staticmethod
    def send(event: int, context: dict, user: User = None, email: str = None, phone: str = None,  # noqa: C901
             files: list = None, data_json: dict = None, viber_chat_id: str = None, room_name: str = None,
             notify_templates: list = None, send_at: datetime = None, send_now: bool = False,
             user_want_message_check: bool = False, **kwargs) -> List[Optional['Notify']]:

        if user and not isinstance(user, User):
            raise IsInstanceException()

        instance_list: list = []

        local_context = context.copy()
        data_json = json.dumps(data_json) if data_json else None

        if notify_templates:
            templates = (
                NotifyTemplate.objects
                .select_related('category', 'user')
                .prefetch_related('user_lists')
                .filter(id__in=notify_templates, event=event, is_active=True)
            )
        else:
            templates = (
                NotifyTemplate.objects
                .select_related('category', 'user')
                .prefetch_related('user_lists')
                .filter(event=event, is_active=True)
            )

        file_instances = []
        if files:
            file_instances = list(map(lambda file: NotifyFile.objects.create(file=file), files))

        for template in templates:

            # Сначала забираем те данные, которые передали с методом
            notify_user = user if user else None
            notify_email = email if email else None
            notify_phone = phone if phone else None
            notify_viber_chat_id = viber_chat_id if viber_chat_id else None

            notify_users_lists = template.user_lists.all()

            # Формируем список основных получателей письма
            # Из шаблона, если методом не было передано данных
            template_user = template.user if template.user else None
            template_email = template.email if template.email else None
            template_phone = template.phone if template.phone else None
            template_viber_chat_id = template.viber_chat_id if template.viber_chat_id else None

            # Данные, которые указал пользователь у себя в профиле в приоритете
            if notify_user:
                notify_email = notify_user.email if notify_user.email else notify_email
                notify_phone = notify_user.phone if notify_user.phone else notify_phone
                notify_viber_chat_id = (
                    notify_user.viber_chat_id if notify_user.viber_chat_id else notify_viber_chat_id
                )

            if notify_user is None and template_user:
                notify_user = template.user if template.user else None
                notify_email = template.user.email if template.user.email else notify_email
                notify_phone = template.user.phone if template.user.phone else notify_phone
                notify_viber_chat_id = (
                    template.user.viber_chat_id if template.user.viber_chat_id else notify_viber_chat_id
                )

            if notify_email is None:
                notify_email = template_email

            if notify_phone is None:
                notify_phone = template_phone

            if notify_viber_chat_id is None:
                notify_viber_chat_id = template_viber_chat_id
            # Проверка, хочет ли пользователь получить сообщение
            if user_want_message_check and hasattr(
                    settings, 'NOTIFY_USER_WANT_MESSAGE_CHECK') and settings.NOTIFY_USER_WANT_MESSAGE_CHECK is not None:

                user_want_message = import_string(settings.NOTIFY_USER_WANT_MESSAGE_CHECK)

                if not notify_users_lists.exists():
                    user_check = user_want_message(event, template.type, notify_user)
                    if not user_check:
                        continue
                else:
                    # Если у нас шаблон со списками, то формируем новый из переданных пользователей,
                    # Если список пустой, то отменяем отправку
                    notify_users_lists = user_want_message(event, template.type, notify_user, notify_users_lists)
                    if not notify_users_lists:
                        continue

            # Передаем пользователя в контекст
            if notify_user is not None:
                if local_context is not None:
                    local_context.update({
                        'user': notify_user,
                    })
                else:
                    local_context = {
                        'user': notify_user,
                    }
            local_context['event_id'] = event

            if send_at is not None:
                notify_send = send_at
            else:
                notify_send = template.send_at

            instance = Notify.objects.create(
                subject=template.render_subject(local_context),
                text=template.render_text(local_context),
                html=template.render_html(local_context),
                user=notify_user,
                email=notify_email,
                phone=notify_phone if notify_phone is not None else "",
                viber_chat_id=notify_viber_chat_id if notify_viber_chat_id is not None else "",
                type=template.type,
                event=template.event,
                category=template.category if template.category else None,
                data_json=data_json,
                send_at=notify_send,
                room_name=room_name,
                is_delete_after=template.is_delete_after,
                **kwargs
            )
            if notify_users_lists.exists():
                instance.users_list.add(*notify_users_lists)
            file_instance = instance.files
            for f in file_instances:
                file_instance.add(f)
            instance.save()

            if send_now:
                transaction.on_commit(lambda: instance.start_send())

            instance_list.append(instance)
        return instance_list

    @staticmethod
    def call(phone: str, user: User = None, url: str = None, **kwargs) -> Optional[str]:
        call_url_type: int = CallClient.get_url_type()

        if user and not isinstance(user, User):
            raise IsInstanceException()

        if user is not None:
            phone = user.phone if user.phone else phone

        if url is not None:
            url = url
        else:
            url = SendData.call_url(call_url_type)

        main_url = url.format(to=phone, **kwargs)
        response_url = requests.get(main_url)
        response_dict: dict = response_url.json()
        value, response = CallClient.get_value_checker(response_dict)

        if value == "OK":
            return '{Code}'.format(**response)
        return None

    def to_log(self, error_text: str) -> None:
        log: NotifyErrorLog = NotifyErrorLog(notify=self, error=error_text)
        log.save()

    def get_format_state(self):
        undefined = '<span style="color:black;">Неизвестный статус</span>'
        status: str = StatusMessage.STATUS.get(self.state, undefined)
        return format_html(status)

    get_format_state.short_description = 'Статус'

    def _delete_notify(self) -> None:
        files = self.files.all()
        if files.exists():
            for file in files:
                file_path = f'{settings.MEDIA_ROOT}/{file.file}'
                if os.path.isfile(file_path):
                    os.remove(file_path)
            files.delete()
        self.delete()

    class Meta:
        verbose_name = 'Уведомление'
        verbose_name_plural = 'Уведомления'


@receiver(post_save, sender=Notify)
def system_post_save(sender, instance, created, **kwargs):
    from garpix_notify.tasks.tasks import send_system_notifications
    if created and instance.type == TYPE.SYSTEM:
        transaction.on_commit(lambda: send_system_notifications.delay(instance.pk))



from smtplib import SMTP_SSL, SMTP

from django.conf import settings
from django.db import models
from django.db.models import Manager
from django.utils.html import format_html
from django.utils.timezone import now, timedelta
from django.db import DatabaseError, ProgrammingError

from .category import NotifyCategory
from .config import NotifyConfig

try:
    config = NotifyConfig.get_solo()
    EMAIL_MAX_HOUR_LIMIT = config.email_max_hour_limit
    EMAIL_MAX_DAY_LIMIT = config.email_max_day_limit
except (DatabaseError, ProgrammingError):
    EMAIL_MAX_HOUR_LIMIT = getattr(settings, 'EMAIL_MAX_HOUR_LIMIT', 240)
    EMAIL_MAX_DAY_LIMIT = getattr(settings, 'EMAIL_MAX_DAY_LIMIT', 240)


class SMTPAccount(models.Model):
    host = models.CharField(default='smtp.yandex.com', max_length=255, verbose_name='Хост')
    port = models.IntegerField(default=465, verbose_name='Порт')
    is_use_tls = models.BooleanField(default=False, verbose_name='Использовать TLS?')
    is_use_ssl = models.BooleanField(default=True, verbose_name='Использовать SSL?')
    sender = models.CharField(max_length=255, blank=True, default='', verbose_name='Отправитель')
    username = models.CharField(max_length=255, blank=True, default='', verbose_name='Имя пользователя')
    password = models.CharField(max_length=255, blank=True, default='', verbose_name='Пароль пользователя')
    timeout = models.IntegerField(default=5000, verbose_name='Таймаут (сек.)')
    category = models.ForeignKey(NotifyCategory, on_delete=models.CASCADE, related_name='accounts', verbose_name='Тип')
    is_active = models.BooleanField(default=True, verbose_name='Включить аккаунт')

    email_hour_used_times = models.IntegerField(default=0, verbose_name='Количество использований в час')
    email_day_used_times = models.IntegerField(default=0, verbose_name='Количество использований в день')
    email_hour_used_date = models.DateTimeField(default=now, verbose_name='Дата использований в час')
    email_day_used_date = models.DateTimeField(default=now, verbose_name='Дата использований в день')

    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Дата создания')

    objects = Manager()

    @classmethod
    def get_free_smtp(cls):
        """
        Проверяет и отдает один аккаунт SMTP, с которого можно отправить письмо
        """
        account_smtp = None

        hour_limit = now() - timedelta(hours=1)
        day_limit = now() - timedelta(days=1)

        accounts = cls.objects.filter(is_active=True).distinct()

        if accounts.exists():

            for account in accounts:

                if account.email_hour_used_date < hour_limit:
                    account.email_hour_used_date = now()
                    account.email_hour_used_times = 0
                elif account.email_hour_used_times > EMAIL_MAX_HOUR_LIMIT:
                    continue

                if account.email_day_used_date < day_limit:
                    account.email_day_used_date = now()
                    account.email_day_used_times = 0
                elif account.email_day_used_times > EMAIL_MAX_DAY_LIMIT:
                    continue

                account.email_hour_used_times += 1
                account.email_day_used_times += 1
                account.save()
                account_smtp = account
                break
        return account_smtp

    def is_worked_now(self):

        hour_limit = now() - timedelta(hours=1)
        day_limit = now() - timedelta(days=1)

        if self.email_day_used_times >= EMAIL_MAX_DAY_LIMIT and self.email_day_used_date > day_limit:
            return format_html('<span style="color:red;">Превышен суточный лимит</span>')
        elif not self.is_active:
            return format_html('<span style="color:red;">Отключен</span>')
        elif self.email_hour_used_times >= EMAIL_MAX_HOUR_LIMIT and self.email_hour_used_date > hour_limit:
            return format_html('<span style="color:orange;">Превышен часовой лимит</span>')
        else:
            return format_html('<span style="color:green;">Включен</span>')

    is_worked_now.short_description = 'Состояние'

    def __str__(self):
        return '{} ({}:{})'.format(self.username, self.host, self.port)

    def clear(self):
        self.email_hour_used_times = 0
        self.email_day_used_times = 0
        self.email_hour_used_date = now()
        self.email_day_used_date = now()
        self.save()

    def test_account(self):
        server = SMTP_SSL(self.host, self.port, timeout=self.timeout) if self.is_use_ssl else SMTP(
            self.host, self.port, timeout=self.timeout)
        server.ehlo()

    class Meta:
        verbose_name = 'SMTP аккаунт'
        verbose_name_plural = 'SMTP аккаунты'



import re

from django.conf import settings
from django.contrib import admin, messages

from ..models import SystemNotify, Notify
from ..models.choices import TYPE
from ..models.template import NotifyTemplate
from django.http import HttpResponseRedirect


@admin.register(NotifyTemplate)
class NotifyTemplateAdmin(admin.ModelAdmin):
    change_form_template = "send_notify.html"
    fields = (
        'title',
        'is_active',
        'subject',
        'is_delete_after',
        'get_context_description',
        'text',
        'html_from_type',
        'html',
        'zipfile',
        'user',
        'email',
        'phone',
        'telegram_chat_id',
        'viber_chat_id',
        'type',
        'category',
        'event',
        'user_lists',
        'send_at',
    )
    readonly_fields = (
        'get_context_description',
        'get_event_description',
    )
    list_display = ('title', 'is_active', 'type', 'category', 'event', 'user', 'email', 'phone', 'send_at')
    list_filter = ('type', 'category', 'event', 'is_active')
    actions = ['create_mailing', ]
    filter_horizontal = ('user_lists',)
    raw_id_fields = ('user',)

    def create_mailing(self, request, queryset):
        count = Notify.send(event=None, context={}, notify_templates=queryset)
        self.message_user(request, 'Рассылка создана, кол-во сообщений: {}'.format(count))

    create_mailing.short_description = "Сделать рассылку"

    def response_change(self, request, obj):
        template = obj
        context = obj.get_test_data()
        user = obj.user if obj.user else request.user

        type_test_message = {
            TYPE.SMS: {
                'subject': obj.render_subject(template.subject),
                'text': obj.render_text(context),
                'user': user,
                'phone': template.phone,
                'type': TYPE.SMS,
                'category': obj.category
            },
            TYPE.EMAIL: {
                'subject': obj.render_subject(template.subject),
                'text': obj.render_text(context),
                'html': obj.render_html(context),
                'user': user,
                'email': template.email,
                'type': TYPE.EMAIL,
                'category': obj.category
            },
            TYPE.PUSH: {
                'subject': obj.render_subject(template.subject),
                'text': obj.render_text(context),
                'user': user,
                'type': TYPE.PUSH,
                'category': obj.category
            },
            TYPE.CALL: {
                'subject': obj.render_subject(template.subject),
                'text': obj.render_text(context),
                'user': user,
                'phone': template.phone,
                'type': TYPE.SMS,
                'category': obj.category
            },
        }

        if obj.user_lists and "_send_now" in request.POST:
            instance = Notify.objects.create(**type_test_message.get(obj.type, TYPE.EMAIL))
            instance.start_send()
            self.message_user(request, 'Тестовое уведомление отправлено', level=messages.SUCCESS)
            return HttpResponseRedirect(".")
        elif "_send_now_system" in request.POST:
            instance = SystemNotify.objects.create(
                title=template.subject if template.subject or template.subject != '' else template.title,
                event=template.event,
                user=user,
                data_json=context,
                room_name=f'room_{user.pk}'
            )
            instance.send_notification()
            self.message_user(request, 'Тестовое уведомление отправлено', level=messages.SUCCESS)
            return HttpResponseRedirect(".")

        return super().response_change(request, obj)

    def get_changelist(self, request, **kwargs):
        events_message = NotifyTemplate.get_blank_events_message()
        if events_message:
            self.message_user(request, events_message, level=messages.WARNING)
        return super().get_changelist(request, **kwargs)

    def save_model(self, request, obj, form, change):
        if hasattr(obj, '_html_file'):
            with open(obj._html_file, 'r') as f:
                _html = f.read()
                for img in obj._images:
                    _full_img_url = request.build_absolute_uri(
                        f"{settings.MEDIA_URL}{obj._secret_path}/{img['file_path']}")
                    _html = re.sub(r"src *= *[\"'](\./)*({0})[\"']".format(img['html_path']), f"src='{_full_img_url}'",
                                   _html)
                    _html = re.sub(r"url *\( *[\"'](\./)*({0})[\"'] *\)".format(img['html_path']),
                                   f"url('{_full_img_url}')", _html)
                obj.html = _html
        super().save_model(request, obj, form, change)

    class Media:
        css = {
            'all': ('css/admin/garpix_notify.css',)
        }
        js = ('js/admin/garpix_notify.js',)



from django.contrib import admin, messages
from django.http import HttpResponseRedirect

from ..models.smtp import SMTPAccount


def clear_limits(modeladmin, request, queryset):
    for server in queryset.all():
        server.clear()


clear_limits.short_description = 'Обнулить лимиты'


@admin.register(SMTPAccount)
class SMTPAccountAdmin(admin.ModelAdmin):
    change_form_template = "test_smtp_account.html"
    list_display = ('__str__', 'is_worked_now')
    readonly_fields = ('email_hour_used_times', 'email_hour_used_date', 'email_day_used_times', 'email_day_used_date')
    actions = [
        clear_limits
    ]

    def response_change(self, request, obj):

        if "_test_smtp" in request.POST:
            try:
                obj.test_account()
                self.message_user(request, 'Успешно', level=messages.SUCCESS)
            except Exception as e:
                self.message_user(request, str(e), level=messages.ERROR)
            return HttpResponseRedirect(".")

        return super().response_change(request, obj)



from django.db import models


class NotifyErrorLog(models.Model):
    notify = models.ForeignKey('garpix_notify.Notify', on_delete=models.CASCADE, related_name='logs',
                               verbose_name='Notify')
    error = models.TextField(verbose_name='Ошибка', blank=True, null=True, default='')
    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Дата создания')

    def __str__(self):
        return '{}: {}'.format(self.pk, self.created_at.strftime('%d.%m.%Y %h:%M:%s'))

    class Meta:
        verbose_name = 'Лог отправки уведомления'
        verbose_name_plural = 'Логи отправки уведомления'



# notify config

PERIODIC_SENDING = 60
EMAIL_MAX_DAY_LIMIT = 240
EMAIL_MAX_HOUR_LIMIT = 240
# SMS
SMS_URL_TYPE = 0
SMS_API_ID = 1234567890
SMS_LOGIN = ''
SMS_PASSWORD = ''
SMS_FROM = ''
# CALL
CALL_URL_TYPE = 0
CALL_API_ID = 1234567890
CALL_LOGIN = ''
CALL_PASSWORD = ''
# TELEGRAM
TELEGRAM_API_KEY = '000000000:AAAAAAAAAA-AAAAAAAA-_AAAAAAAAAAAAAA'
TELEGRAM_BOT_NAME = 'MySuperBot'
TELEGRAM_WELCOME_TEXT = 'Hello'
TELEGRAM_HELP_TEXT = '/set !help for HELP'
TELEGRAM_BAD_COMMAND_TEXT = 'Incorrect command format'
TELEGRAM_SUCCESS_ADDED_TEXT = 'Success'
TELEGRAM_FAILED_ADDED_TEXT = 'Failed'
TELEGRAM_PARSE_MODE = None
TELEGRAM_DISABLE_NOTIFICATION = False
TELEGRAM_DISABLE_PAGE_PREVIEW = False
TELEGRAM_SENDING_WITHOUT_REPLY = False
TELEGRAM_TIMEOUT = None
# VIBER
VIBER_API_KEY = '000000000:AAAAAAAAAA-AAAAAAAA-_AAAAAAAAAAAAAA'
VIBER_BOT_NAME = 'MySuperViberBot'
VIBER_WELCOME_TEXT = 'Hello'
VIBER_SUCCESS_ADDED_TEXT = 'Success'
VIBER_FAILED_ADDED_TEXT = 'Failed'
VIBER_TEXT_FOR_NEW_SUB = 'HI!'
# WHATSAPP
IS_WHATS_APP_ENABLED = True
WHATS_APP_AUTH_TOKEN = None
WHATS_APP_ACCOUNT_SID = None
WHATS_APP_NUMBER_SENDER = None
# SETTINGS
EMAIL_MALLING = 1

NOTIFY_USER_WANT_MESSAGE_CHECK = None
NOTIFY_CALL_CODE_CHECK = None
GARPIX_NOTIFY_CELERY_SETTINGS = 'app.celery.app'
DEFAULT_SYSTEM_NOTIFY_TYPE = 'system'



from django.utils.crypto import get_random_string
from garpix_notify.models.choices import TYPE

some_text = 'bla bla'


def generate_users(count: int) -> list:
    users_list = [{
        'username': f'user_{i}',
        'email': f'{get_random_string(length=4)}@garpix.com',
        'password': f'BlaBla123{get_random_string(length=4)}',
        'first_name': f'{get_random_string(length=5)}',
        'last_name': f'{get_random_string(length=8)}',
    } for i in range(count)]
    return users_list


def generate_viber_user(count: int) -> list:
    users_list = [{
        'username': 'viber_' + get_random_string(length=5),
        'email': f'{get_random_string(length=4)}@garpix.com',
        'viber_secret_key': get_random_string(length=5, allowed_chars='1234567890'),
        'viber_chat_id': 'm4FsaRu5kBi8HzSAC0liFQ==',
        'password': f'BlaBla123{get_random_string(length=4)}',
        'first_name': f'{get_random_string(length=5)}',
        'last_name': f'{get_random_string(length=8)}',
    } for i in range(count)]
    return users_list


def generate_templates(event: int) -> list:
    return [{
        'title': 'Тестовый темплейт',
        'subject': 'Тестовый темплейт {{user.email}}',
        'text': 'Контент текстовый {{user.email}} - {{sometext}}',
        'html': 'Контент HTML {{user.email}} - {{sometext}}',
        'type': TYPE.EMAIL,
        'event': event,
    }]


def generate_templates_viber(event: int) -> list:
    return [{
        'title': 'Тестовый темплейт вайбер',
        'subject': 'Тестовый темплейт {{user.viber_chat_id}}',
        'text': 'Контент текстовый {{user.viber_chat_id}} - {{sometext}}',
        'html': 'Контент HTML {{user.viber_chat_id}} - {{sometext}}',
        'type': TYPE.VIBER,
        'event': event,
    }]


def generate_category() -> list:
    return [{
        'title': 'Основная категория_' + get_random_string(length=4),
        'template': '<div>{{text}}</div>',
    }]


def generate_compiled_email(data_user: dict, event: int) -> list:
    return [{
        'title': 'Тестовый темплейт',
        'subject': f'Тестовый темплейт {data_user["email"]}',
        'text': f'Контент текстовый {data_user["email"]} - {some_text}',
        'html': f'Контент HTML {data_user["email"]} - {some_text}',
        'type': TYPE.EMAIL,
        'event': event,
    }]


def generate_compiled_viber(viber_chat_id: str, event: int) -> list:
    return [{
        'title': 'Тестовый темплейт вайбер',
        'subject': f'Тестовый темплейт {viber_chat_id}',
        'text': f'Контент текстовый {viber_chat_id} - {some_text}',
        'html': f'Контент HTML {viber_chat_id} - {some_text}',
        'type': TYPE.VIBER,
        'event': event,
    }]


def generate_notify_data(user) -> dict:
    test_data = {
        'sometext': some_text,
        'user': user
    }
    return test_data


def generate_system_notify_data() -> dict:
    return {f"{get_random_string(length=4)}": f"{get_random_string(length=7)}"}


def generate_system_template_data(value: dict, event: int) -> list:
    return [{
        'title': 'Тестовый системный темплейт',
        'subject': 'Тестовый темплейт',
        'text': value,
        'html': value,
        'type': TYPE.SYSTEM,
        'event': event,
    }]


def generate_compiled_system(value: dict, event: int, user) -> list:
    value['type'] = 'system'
    value['user'] = user.pk
    return [{
        "title": 'Тестовый темплейт',
        "data_json": value,
        "type": TYPE.SYSTEM,
        "event": event,
    }]


def generate_templates_views(event: int) -> list:
    return [{
        'title': 'Тестовый темплейт',
        'subject': 'Тестовый темплейт',
        'text': '{{message}} - {{confirmation_code}}',
        'html': '{{message}} - {{confirmation_code}}',
        'type': TYPE.EMAIL,
        'event': event,
    }]


def generate_compiled_email_views(event: int) -> list:
    return [{
        'title': 'Тестовый темплейт',
        'subject': 'Тестовый темплейт',
        'text': 'Привет, тут твое сообщение - abcdef12345',
        'html': 'Привет, тут твое сообщение - abcdef12345',
        'email': 'example@garpix.com',
        'type': TYPE.EMAIL,
        'event': event,
    }]



from django.conf import settings
from django.urls import path
from .views import send_webhook, viber_check_webhook, SystemNotifyViewSet
from rest_framework.routers import DefaultRouter

app_name = 'garpix_notify'

router = DefaultRouter()

router.register(f'{settings.API_URL}/garpix_notify/system_notifies', SystemNotifyViewSet, basename='system_notifies')

urlpatterns = [
    path('send_webhook', send_webhook, name='send_webhook'),
    path('viber_check_webhook', viber_check_webhook, name='viber_check_webhook'),
] + router.urls



import pytest

from garpix_notify.models import Notify
from .utils.generate_data import (
    generate_users, generate_category, generate_templates, generate_compiled_email, generate_notify_data,
    generate_viber_user, generate_compiled_viber, generate_templates_viber
)
from .utils.common_class import CommonTestClass


@pytest.mark.django_db
class TestNotify(CommonTestClass):

    @pytest.fixture
    def setup(self):
        # Эвенты для теста
        self.PASS_TEST_EVENT = 0
        self.PASS_TEST_EVENT_1 = 1
        self.PASS_TEST_EVENT_2 = 2
        # Тестовый пользователь
        self.data_user = generate_users(3)[2]
        self.data_viber_user = generate_viber_user(2)[1]
        self.user_1 = self.create_user(self.data_user)
        self.user_viber = self.create_user(self.data_viber_user)

        # Общие данные
        self.user_list = self.create_user_list()
        self.data_category = generate_category()[0]
        self.viber_chat_id = 'm4FsaRu5kBi8HzSAC0liFQ=='
        self.test_data = generate_notify_data(self.user_1)

        # Генерация данных для создания шаблонов и данных для теста
        self.data_template_email = generate_templates(self.PASS_TEST_EVENT)[0]
        self.data_compiled_email = generate_compiled_email(self.data_user, self.PASS_TEST_EVENT)[0]

        self.data_template_email_2 = generate_templates(self.PASS_TEST_EVENT_1)[0]
        self.data_compiled_email_2 = generate_compiled_email(self.data_user, self.PASS_TEST_EVENT_1)[0]

        self.data_template_viber = generate_templates_viber(self.PASS_TEST_EVENT_2)[0]
        self.data_compiled_viber = generate_compiled_viber(self.viber_chat_id, self.PASS_TEST_EVENT_2)[0]

        # Создание шаблонов/категорий
        self.category_1 = self.create_category(self.data_category)
        self.template_1 = self.create_template(self.data_template_email, self.category_1)
        self.template_2 = self.create_template(self.data_template_email_2, self.category_1, self.user_list)
        self.template_viber = self.create_template(self.data_template_viber, self.category_1, self.user_list)

    def test_notify_email_positive(self, setup):
        # Сверяем корректно ли создались объекты
        assert self.category_1.title == self.data_category['title']
        assert self.category_1.template == self.data_category['template']
        assert self.template_1.title == self.data_template_email['title']
        assert self.template_1.subject, self.data_template_email['subject']
        assert self.template_1.text == self.data_template_email['text']
        assert self.template_1.html == self.data_template_email['html']
        assert self.template_1.type == self.data_template_email['type']
        assert self.template_1.event == self.data_template_email['event']

        # Создаем письмо
        notify = Notify.send(event=self.PASS_TEST_EVENT, context=self.test_data, user=self.user_1)

        # Сверяем с шаблоном и ожидаемым результатом
        assert notify[0].event == self.template_1.event
        assert notify[0].subject == self.data_compiled_email['subject']
        assert notify[0].text == self.data_compiled_email['text']
        assert notify[0].html == self.data_compiled_email['html']
        assert notify[0].type == self.data_compiled_email['type']
        assert notify[0].event == self.data_compiled_email['event']

    def test_notify_email_user_list(self, setup):
        # Сверяем корректно ли создались объекты
        assert self.template_2.title == self.data_template_email_2['title']
        assert self.template_2.subject, self.data_template_email_2['subject']
        assert self.template_2.text == self.data_template_email_2['text']
        assert self.template_2.html == self.data_template_email_2['html']
        assert self.template_2.type == self.data_template_email_2['type']
        assert self.template_2.event == self.data_template_email_2['event']
        assert self.template_2.category.id == self.category_1.id

        # Создаем письмо
        notify = Notify.send(event=self.PASS_TEST_EVENT_1, context=self.test_data, user=self.user_1)

        # Сверяем с шаблоном и ожидаемым результатом
        assert notify[0].event == self.template_2.event
        assert notify[0].email == self.user_1.email
        assert notify[0].subject == self.data_compiled_email_2['subject']
        assert notify[0].text == self.data_compiled_email_2['text']
        assert notify[0].html == self.data_compiled_email_2['html']
        assert notify[0].type == self.data_compiled_email_2['type']
        assert notify[0].event == self.data_compiled_email_2['event']

    def test_notify_viber(self, setup):
        # Сверяем корректно ли создались объекты
        assert self.template_viber.title == self.data_template_viber['title']
        assert self.template_viber.subject, self.data_template_viber['subject']
        assert self.template_viber.text == self.data_template_viber['text']
        assert self.template_viber.html == self.data_template_viber['html']
        assert self.template_viber.type == self.data_template_viber['type']
        assert self.template_viber.event == self.data_template_viber['event']
        assert self.template_viber.category.id == self.category_1.id

        # Создаем письмо
        notify = Notify.send(event=self.PASS_TEST_EVENT_2, context=self.test_data, user=self.user_viber)

        # Сверяем с шаблоном и ожидаемым результатом
        assert notify[0].event == self.template_viber.event
        assert notify[0].email == self.user_viber.email
        assert notify[0].subject == self.data_compiled_viber['subject']
        assert notify[0].text == self.data_compiled_viber['text']
        assert notify[0].html == self.data_compiled_viber['html']
        assert notify[0].type == self.data_compiled_viber['type']
        assert notify[0].event == self.data_compiled_viber['event']
        assert notify[0].user.viber_chat_id == self.viber_chat_id



from .category import NotifyCategoryAdmin  # noqa
from .config import NotifyConfigAdmin  # noqa
from .notify import NotifyAdmin  # noqa
from .smtp import SMTPAccountAdmin  # noqa
from .fcm import NotifyDeviceAdmin  # noqa
from .template import NotifyTemplateAdmin  # noqa
from .file import NotifyFileAdmin  # noqa
from .user_list import NotifyUserListAdmin  # noqa
from .system_log import SystemNotifyErrorLogInline  # noqa
from .system_notify import SystemNotifyAdmin  # noqa



import requests
import enum

from typing import Optional, Tuple
from django.conf import settings
from django.db import DatabaseError, ProgrammingError
from django.utils.timezone import now

from garpix_notify.models.config import NotifyConfig
from garpix_notify.utils.send_data import SendData
from garpix_notify.models.choices import STATE, CALL_URL


class CallClient:

    class ChoiceValue(enum.Enum):
        OK = 0
        BAD = 1

    def __init__(self, notify):
        self.notify = notify
        try:
            self.config = NotifyConfig.get_solo()
            self.IS_CALL_ENABLED = self.config.is_call_enabled
            self.CALL_URL_TYPE = self.config.call_url_type
        except (DatabaseError, ProgrammingError):
            self.IS_CALL_ENABLED = getattr(settings, 'IS_CALL_ENABLED', True)
            self.CALL_URL_TYPE = getattr(settings, 'CALL_URL_TYPE', 0)

    def __value_checker(self, response_dict: dict) -> Optional[int]:
        value: Optional[int] = None

        if self.CALL_URL_TYPE in (CALL_URL.SMSRU_CALL_ID, CALL_URL.SMSRU_CALL_API_ID):
            if response_dict['status'] == "OK":
                value = self.ChoiceValue.OK.value
            else:
                value = self.ChoiceValue.BAD.value

        elif self.CALL_URL_TYPE == CALL_URL.SMSCENTRE_ID:
            if response_dict['error']:
                value = self.ChoiceValue.BAD.value
            else:
                value = self.ChoiceValue.OK.value

        elif self.CALL_URL_TYPE == CALL_URL.UCALLER_ID:
            if response_dict['status']:
                value = self.ChoiceValue.OK.value
            else:
                value = self.ChoiceValue.BAD.value
        return value

    def __response_check(self, response: dict, status: int) -> dict:
        if status == self.ChoiceValue.OK.value:
            response_processing: dict = {
                0: {'Status': response.get('status'),
                    'Code': response.get('code'),
                    'Balance': response.get('balance'),
                    'ID_Call': response.get('call_id')},
                1: {'Status': response.get('status'),
                    'Code': response.get('code'),
                    'Balance': response.get('balance'),
                    'ID_Call': response.get('call_id')},

                2: {'Status': response.get('id'),
                    'Code': response.get('code'),
                    'ID_Call': response.get('cnt'),
                    'Balance': response.get('balance')},

                3: {'Status': response.get('status'),
                    'Code': response.get('code'),
                    'Balance': response.get('balance'),
                    'ID_Call': response.get('unique_request_id')}}

        else:
            response_processing: dict = {
                0: {'Status': response.get('status'),
                    'Status_code': response.get('status_code'),
                    'Status_text': response.get('status_text')},
                1: {'Status': response.get('status'),
                    'Status_code': response.get('status_code'),
                    'Status_text': response.get('status_text')},
                2: {'Status': response.get('status'),
                    'Status_code': response.get('error_code'),
                    'Status_text': response.get('error')},
                3: {'Status': response.get('status'),
                    'Status_code': response.get('code'),
                    'Status_text': response.get('error')}}

        return response_processing[self.CALL_URL_TYPE]

    def __send_call_code(self) -> None:

        if not self.IS_CALL_ENABLED:
            self.notify.state = STATE.DISABLED
            self.notify.to_log('Not sent (sending is prohibited by settings)')
            return

        phone = str(self.notify.phone)

        try:
            send_data = SendData()
            url = send_data.call_url(self.CALL_URL_TYPE).format(to=phone)
            response_url = requests.get(url)
            response_dict = response_url.json()
            value = self.__value_checker(response_dict)
            response = self.__response_check(response_dict, value)
            self.__save_to_log(response, value)
        except Exception as e:
            self.notify.state = STATE.REJECTED
            self.notify.to_log(str(e))

    def __save_to_log(self, response: dict, value: int) -> None:
        if value == self.ChoiceValue.OK.value:
            self.notify.to_log(
                'Status: {Status}, Code: {Code}, Balance: {Balance}, ID_Call: {ID_Call}'.format(**response)
            )
            self.notify.state = STATE.DELIVERED
            self.notify.sent_at = now()
        elif value == self.ChoiceValue.BAD.value:
            self.notify.to_log(
                'Status: {Status}, Status_code: {Status_code}, Status_text: {Status_text}'.format(**response)
            )
            self.notify.state = STATE.REJECTED

    @classmethod
    def get_value_checker(cls, response_dict: dict) -> Tuple[Optional[int], dict]:
        call_client_empty = cls(notify=None)
        value = call_client_empty.__value_checker(response_dict)
        return value, call_client_empty.__response_check(response_dict, value)

    @classmethod
    def get_url_type(cls) -> int:
        return cls(notify=None).CALL_URL_TYPE

    @classmethod
    def send_call(cls, notify) -> None:
        cls(notify).__send_call_code()



from django.contrib import admin
from ..models.system_log import SystemNotifyErrorLog


class SystemNotifyErrorLogInline(admin.TabularInline):
    model = SystemNotifyErrorLog
    fields = ('created_at', 'error')
    readonly_fields = ('created_at', 'error')
    extra = 0

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False



from django.core.mail.backends.smtp import EmailBackend as DjangoEmailBackend
from garpix_notify.models.smtp import SMTPAccount

try:
    smtp = SMTPAccount.get_free_smtp()
except Exception as e:
    raise ("[SMTPAccount not found] Проверьте настроки в админке : /admin/garpix_notify/smtpaccount/", e)


class EmailBackend(DjangoEmailBackend):

    def __init__(self, host=smtp.host, port=smtp.port, username=smtp.sender, password=smtp.password,
                 use_tls=smtp.is_use_tls, fail_silently=False, use_ssl=smtp.is_use_ssl, timeout=None,
                 ssl_keyfile=None, ssl_certfile=None, **kwargs):
        super().__init__(host, port, username, password, use_tls, fail_silently, use_ssl, timeout,
                         ssl_keyfile, ssl_certfile, **kwargs)



from smtplib import SMTP, SMTP_SSL

from email.header import Header
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication

from django.conf import settings
from django.utils.timezone import now
from django.utils.safestring import mark_safe
from django.template import Template, Context
from django.db import DatabaseError, ProgrammingError

from garpix_notify.models.choices import STATE, EMAIL_MALLING
from garpix_notify.models.category import NotifyCategory
from garpix_notify.models.config import NotifyConfig
from garpix_notify.models.smtp import SMTPAccount
from garpix_notify.utils import ReceivingUsers


class EmailClient:

    def __init__(self, notify):
        self.emails: list = []
        self.notify = notify
        try:
            self.config = NotifyConfig.get_solo()
            self.IS_EMAIL_ENABLED = self.config.is_email_enabled
            self.EMAIL_MALLING_TYPE = self.config.email_malling
        except (DatabaseError, ProgrammingError):
            self.IS_EMAIL_ENABLED = getattr(settings, 'IS_EMAIL_ENABLED', True)
            self.EMAIL_MALLING_TYPE = getattr(settings, 'EMAIL_MALLING', EMAIL_MALLING.BCC)

    def __render_body(self, mail_from: str, layout: NotifyCategory, emails: list) -> MIMEMultipart:
        # =============================== Основная часть письма ===========================================
        msg = MIMEMultipart('mixed')
        if len(emails) > 1:
            if self.EMAIL_MALLING_TYPE == EMAIL_MALLING.BCC:
                msg['BCC'] = ', '.join(emails)
            else:
                msg['СС'] = ', '.join(emails)
        else:
            msg['To'] = ''.join(emails)

        msg['Subject'] = Header(self.notify.subject, 'utf-8')
        msg['From'] = mail_from

        if self.notify.files.exists():
            for fl in self.notify.files.all():
                file_name = fl.file.name.split('/')[-1]
                with fl.file.open(mode='rb') as f:
                    part = MIMEApplication(
                        f.read(),
                        Name=file_name
                    )

                part['Content-Disposition'] = 'attachment; filename="%s"' % file_name

                msg.attach(part)

        # =============================== Часть письма с текстом ==========================================
        msg_text = MIMEMultipart('alternative')
        text = MIMEText(self.notify.text, 'plain', 'utf-8')
        msg_text.attach(text)

        if self.notify.html:
            template = Template(layout.template)
            context = Context({'text': mark_safe(self.notify.html)})
            html = MIMEText(mark_safe(template.render(context)), 'html', 'utf-8')
            msg_text.attach(html)

        msg.attach(msg_text)

        return msg

    def __send_email_client(self) -> None:

        if not self.IS_EMAIL_ENABLED:
            self.notify.state = STATE.DISABLED
            self.notify.to_log('Not sent (sending is prohibited by settings)')
            return
        account = SMTPAccount.get_free_smtp()

        if account is None:
            self.notify.state = STATE.DISABLED
            self.notify.to_log('No SMTPAccount')
            return

        if self.notify.sender_email is None:
            self.notify.sender_email = account.sender

        users_list = self.notify.users_list.all().order_by('-mail_to_all')

        if users_list.exists():
            self.emails: list = ReceivingUsers.run_receiving_users(users_list, 'email')

        if self.notify.email:
            self.emails.append(self.notify.email)

        try:
            body = self.__render_body(mail_from=account.sender, layout=self.notify.category, emails=self.emails)
            server = SMTP_SSL(account.host, account.port, timeout=account.timeout) if account.is_use_ssl else SMTP(account.host, account.port, timeout=account.timeout)
            server.ehlo()
            if account.is_use_tls:
                server.starttls()
            server.login(account.username, account.password)
            server.sendmail(account.sender, self.emails, body.as_string())
            server.close()
            self.notify.state = STATE.DELIVERED
            self.notify.sent_at = now()
        except Exception as e:
            self.notify.state = STATE.REJECTED
            self.notify.to_log(str(e))

    @classmethod
    def send_email(cls, notify) -> None:
        cls(notify).__send_email_client()



from django.contrib import admin

from ..models.system_notify import SystemNotify
from .system_log import SystemNotifyErrorLogInline


@admin.register(SystemNotify)
class SystemNotifyAdmin(admin.ModelAdmin):
    list_display = ('__str__', 'type', 'event', 'user', 'get_format_state', 'created_at', 'sent_at')
    readonly_fields = ('created_at', 'type')
    list_filter = ('event', 'state')
    search_fields = ('title', 'user__username')
    raw_id_fields = ('user',)
    inlines = [SystemNotifyErrorLogInline]




from django.core.exceptions import ValidationError
from django.core.validators import (
    FileExtensionValidator,
)


def validate_file_extension(value, allowed_extensions):
    return FileExtensionValidator(allowed_extensions=allowed_extensions)(value)


def validate_zip(value):
    allowed_extensions = ['zip', 'rar']
    validate_file_extension(value, allowed_extensions)
    filesize = value.file.size
    megabyte_limit = 100.0
    if filesize > megabyte_limit * 1024 * 1024:
        raise ValidationError("Максимальный допустимый размер изображения %sMB" % str(megabyte_limit))


def validate_zip_files(value):
    allowed_extensions = ['png', 'jpg', 'jpeg', 'html']
    validate_file_extension(value, allowed_extensions)



import json

from django.conf import settings
from django.utils.timezone import now
from django.db import DatabaseError, ProgrammingError

from garpix_notify.models.config import NotifyConfig
from garpix_notify.models.fcm import NotifyDevice
from garpix_notify.models.choices import STATE


class PushClient:

    def __init__(self, notify):
        self.notify = notify
        try:
            self.config = NotifyConfig.get_solo()
            self.IS_PUSH_ENABLED = self.config.is_push_enabled
        except (DatabaseError, ProgrammingError):
            self.IS_PUSH_ENABLED = getattr(settings, 'IS_PUSH_ENABLED', True)

    def __send_push_client(self):

        if not self.IS_PUSH_ENABLED or self.notify.user is None:
            if self.notify.user is None:
                message = 'No users found'
                self.notify.state = STATE.REJECTED
            else:
                message = 'Not sent (sending is prohibited by settings)'
                self.notify.state = STATE.DISABLED
            self.notify.to_log(message)
            return

        devices = NotifyDevice.objects.filter(active=True, user=self.notify.user).distinct('device_id')
        data_json: dict = {}

        try:
            data_json.update({
                'notify_id': self.notify.id,
                'sound': 'default'
            })

            if self.notify.data_json is not None:
                data_json.update(json.loads(self.notify.data_json))

            devices.send_message(title=self.notify.subject, body=self.notify.text, data=data_json)
            self.notify.state = STATE.DELIVERED
            self.notify.sent_at = now()
        except Exception as e:
            self.notify.state = STATE.REJECTED
            self.notify.to_log(str(e))

    @classmethod
    def send_push(cls, notify):
        cls(notify).__send_push_client()



class TYPE:
    EMAIL = 0
    SMS = 1
    PUSH = 2
    TELEGRAM = 3
    VIBER = 4
    SYSTEM = 5
    CALL = 6
    WHATSAPP = 7

    CHOICES = (
        (EMAIL, 'E-mail'),
        (SMS, 'SMS'),
        (PUSH, 'Push'),
        (TELEGRAM, 'Telegram'),
        (VIBER, 'Viber'),
        (SYSTEM, 'System'),
        (CALL, 'Call'),
        (WHATSAPP, 'WhatsApp')
    )


class STATE:
    DISABLED = -2
    REJECTED = -1
    WAIT = 0
    DELIVERED = 1

    CHOICES = (
        (DELIVERED, 'Доставлено'),
        (REJECTED, 'Отклонено'),
        (WAIT, 'В ожидании'),
        (DISABLED, 'Не отправлено (отправка запрещена настройками)'),
    )


class EMAIL_MALLING:
    CC = 0
    BCC = 1
    TYPES = (
        (CC, 'Обычная рассылка'),
        (BCC, 'Скрытая рассылка'),
    )


class PARSE_MODE_TELEGRAM:
    """ Выбор метода парсинга телеграмм сообщений"""

    EMPTY = ''
    HTML = 'HTML'
    MARKDOWN = 'Markdown'

    TYPES = (
        (EMPTY, 'Без форматирования'),
        (HTML, 'HTML'),
        (MARKDOWN, 'Markdown'),
    )


class SMS_URL:
    """ URL СМС провайдера """

    SMSRU_ID = 0
    WEBSZK_ID = 1
    IQSMS_ID = 2
    INFOSMS_ID = 3
    SMSCENTRE_ID = 4
    SMS_SENDING_ID = 5
    SMS_PROSTO_ID = 6

    SMSRU_URL = 'https://sms.ru/sms/send/'
    WEBSZK_URL = 'https://gateway.api.sc/rest/Send/SendSms/'
    IQSMS_URL = 'https://api.iqsms.ru/messages/v2/send.json'
    INFOSMS_URL = 'http://api.infosmska.ru/interfaces/SendMessages.ashx'
    SMSCENTRE_URL = 'https://smsc.ru/sys/send.php'
    SMS_SENDING_URL = 'http://lcab.sms-sending.ru/lcabApi/sendSms.php'
    SMS_PROSTO_URL = 'http://api.sms-prosto.ru/'

    TYPES = (
        (SMSRU_ID, 'sms.ru'),
        (WEBSZK_ID, 'web.szk-info.ru'),
        (IQSMS_ID, 'iqsms.ru'),
        (INFOSMS_ID, 'infosmska.ru'),
        (SMSCENTRE_ID, 'smsc.ru'),
        (SMS_SENDING_ID, 'sms-sending.ru'),
        (SMS_PROSTO_ID, 'sms-prosto.ru')
    )


class CALL_URL:
    """ URL Оператора связи """
    SMSRU_CALL_API_ID = 0
    SMSRU_CALL_ID = 1
    SMSCENTRE_ID = 2
    UCALLER_ID = 3

    SMSRU_CALL_URL = 'https://sms.ru/code/call'
    SMSCENTRE_URL = 'https://smsc.ru/sys/send.php'
    UCALLER_URL = 'https://api.ucaller.ru/v1.0/initCall'

    TYPES = (
        (SMSRU_CALL_API_ID, 'sms.ru API'),
        (SMSRU_CALL_ID, 'sms.ru LOGIN'),
        (SMSCENTRE_ID, 'smsc.ru'),
        (UCALLER_ID, 'ucaller.ru'),

    )


class StatusMessage:
    """ Статус сообщения """
    STATUS = {
        STATE.DISABLED: '<span style="color:red;">Отправка запрещена</span>',
        STATE.REJECTED: '<span style="color:red;">Отклонено</span>',
        STATE.WAIT: '<span style="color:orange;">В ожидании</span>',
        STATE.DELIVERED: '<span style="color:green;">Отправлено</span>',
    }



from django.conf import settings
from django.contrib.auth.models import Group
from django.db import models
from django.db.models import Manager


class NotifyUserList(models.Model):
    title = models.CharField(max_length=255, verbose_name='Название списка пользователей')
    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Дата создания')
    user_groups = models.ManyToManyField(Group, blank=True, verbose_name='Группы пользователей')
    users = models.ManyToManyField(settings.AUTH_USER_MODEL, blank=True, verbose_name='Пользователи')
    mail_to_all = models.BooleanField(default=False, verbose_name='Массовая рассылка для всех пользователей сайта')

    objects = Manager()

    def __str__(self):
        return self.title

    class Meta:
        verbose_name = 'Список пользователей для рассылки'
        verbose_name_plural = 'Списки пользователей для рассылки'



class ArgumentsEmptyException(Exception):
    message = 'You must pass one of the "user" or "event" or "templates" arguments'

    def __init__(self):
        super().__init__(self.message)


class DataTypeException(Exception):

    def __init__(self, field, data_type):
        self.field = field
        self.data_type = data_type
        message = f'The data type error: "{self.field}" type is not {self.data_type}'
        super().__init__(message)


class IsInstanceException(Exception):
    message = 'The "user" argument must be a "User" instance'

    def __init__(self):
        super().__init__(self.message)


class TemplatesNotExists(Exception):
    message = 'NotifyTemplates not exists'

    def __init__(self):
        super().__init__(self.message)


class UsersListIsNone(Exception):
    message = 'The list of users is empty'

    def __init__(self):
        super().__init__(self.message)



import os
import shutil
from pathlib import Path
from typing import Optional, List

from django.core.exceptions import ValidationError
from django.db import models
from django.conf import settings
from django.db.models import Manager
from django.template import Template, Context

from .category import NotifyCategory
from .choices import TYPE
from .user_list import NotifyUserList

from ckeditor_uploader.fields import RichTextUploadingField
from ..mixins import UserNotifyMixin
from ..utils import get_file_path
from ..utils.validators import validate_zip, validate_zip_files
import zipfile
import rarfile
from garpix_utils.file import get_secret_path


class NotifyTemplate(UserNotifyMixin):
    class HTMLFormType(models.TextChoices):
        CKEDITOR = ('ckeditor', 'Визуальный редактор')
        ZIPFILE = ('zipfile', 'Загрузка из архива')

    title = models.CharField(max_length=255, verbose_name='Название для админа')
    subject = models.CharField(max_length=255, default='', blank=True, verbose_name='Заголовок')
    is_delete_after = models.BooleanField(default=False, verbose_name='Удалять после отправки')
    text = models.TextField(verbose_name='Текст')

    html_from_type = models.TextField(max_length=8, choices=HTMLFormType.choices, default=HTMLFormType.CKEDITOR,
                                      verbose_name='Способ формирования html')
    html = RichTextUploadingField(verbose_name='HTML', blank=True)
    zipfile = models.FileField(upload_to=get_file_path, blank=True, null=True, validators=[validate_zip],
                               verbose_name='Файл архива',
                               help_text="Поддерживаются архивы в форматах ZIP и RAR с единственным HTML-файлом и ассетами изображений по относительному или абсолютному пути")

    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.SET_NULL, blank=True, null=True,
                             verbose_name='Пользователь (получатель)')
    email = models.EmailField(max_length=255, blank=True, null=True, verbose_name='Email получатель',
                              help_text='Используется только в случае отсутствия указанного пользователя')

    type = models.IntegerField(choices=TYPE.CHOICES, verbose_name='Тип')
    category = models.ForeignKey(NotifyCategory, on_delete=models.CASCADE, related_name='templates',
                                 verbose_name='Категория')
    event = models.IntegerField(choices=settings.CHOICES_NOTIFY_EVENT, blank=True, null=True, verbose_name='Событие')

    created_at = models.DateTimeField(auto_now_add=True, verbose_name='Дата создания')
    is_active = models.BooleanField(default=True, verbose_name='Активный')
    send_at = models.DateTimeField(blank=True, null=True, verbose_name='Время начала отправки')

    user_lists = models.ManyToManyField(NotifyUserList, blank=True, verbose_name='Списки пользователей для рассылки')

    objects = Manager()

    def __str__(self):
        return self.title

    def render_subject(self, ct):
        template = Template(self.subject)
        context = Context(ct)
        return template.render(context)

    def render_text(self, ct):
        template = Template(self.text)
        context = Context(ct)
        return template.render(context)

    def render_html(self, ct):
        template = Template(self.html)
        context = Context(ct)
        return template.render(context)

    def get_event_data(self):
        return settings.NOTIFY_EVENTS[self.event]

    def get_context_description(self):
        try:
            text = self.get_event_data()['context_description'] if self.event else '--'
        except:  # noqa
            text = '--'
        return text

    get_context_description.short_description = 'Описание контекста шаблона'

    def get_event_description(self):
        try:
            text = self.get_event_data()['event_description'] if self.event else '--'
        except:  # noqa
            text = '--'
        return text

    get_event_description.short_description = 'Описание события'

    def get_test_data(self):
        try:
            data = self.get_event_data()['test_data'] if self.event else {}
        except:  # noqa
            data = {}
        return data

    @staticmethod
    def get_blank_events_message() -> Optional[str]:
        """ Метод возвращает сообщение о том, что есть Ивенты без шаблонов. """
        message: Optional[str] = None
        events = settings.NOTIFY_EVENTS.keys()
        notify_templates_events = NotifyTemplate.objects.filter(is_active=True).values_list('event', flat=True)
        events_without_templates: List[Optional[int]] = [
            settings.NOTIFY_EVENTS.get(event_id) for event_id in events if event_id not in notify_templates_events
        ]
        if events_without_templates:
            names_events_without_templates: str = ', '.join(
                map(lambda event: event.get('title'), events_without_templates)
            )
            message = 'Не найдено активных шаблонов для ивентов: ' + names_events_without_templates
        return message

    def _parse_and_validate_zipfile(self):

        def _parse_dir(dir_path, imgs, current_folder, root_folders: list):
            for _name in os.listdir(dir_path):
                file_path = os.path.join(dir_path, _name)
                if _name not in ['__MACOSX', '.DS_Store']:
                    if os.path.isdir(file_path):
                        _parse_dir(file_path, imgs, os.path.join(current_folder, _name), root_folders)
                    else:
                        with open(file_path, 'r') as f:
                            validate_zip_files(f)
                            if Path(f.name).suffix[1:].lower() not in ['html', 'htm']:
                                html_path = current_folder
                                if any(root_folders):
                                    _current_folders = current_folder.split('/')
                                    _ci = 0
                                    for _item in root_folders:
                                        if len(_current_folders) == _ci or _current_folders[_ci] != _item:
                                            _current_folders.insert(_ci, '..')
                                        else:
                                            _current_folders[_ci] = None
                                        _ci += 1
                                    html_path = '/'.join([_item for _item in _current_folders if _item is not None])

                                rel_file_path = file_path.split('/')[-1]
                                imgs.append({
                                    'html_path': os.path.join(html_path, rel_file_path),
                                    'file_path': os.path.join(current_folder, rel_file_path)
                                })

        def _validate_and_find_html(dir_path, html_path, root_folder, current_folder) -> (str, str):
            for _name in os.listdir(dir_path):
                file_path = os.path.join(dir_path, _name)
                if _name not in ['__MACOSX', '.DS_Store']:
                    if os.path.isdir(file_path):
                        html_path, root_folder = _validate_and_find_html(file_path, html_path, root_folder,
                                                                         os.path.join(current_folder, _name))
                    else:
                        with open(file_path, 'r') as f:
                            validate_zip_files(f)
                            if Path(f.name).suffix[1:].lower() in ['html', 'htm']:
                                if html_path:
                                    raise ValidationError(
                                        {'zipfile': 'Архив должен содержать только один html файл'})
                                html_path = file_path
                                root_folder = current_folder
            return html_path, root_folder

        try:
            archive = zipfile.ZipFile(self.zipfile, 'r')
        except zipfile.BadZipFile:
            try:
                archive = rarfile.RarFile(self.zipfile)
            except (rarfile.BadRarFile, rarfile.NotRarFile) as e:
                raise ValidationError(
                    {'zipfile': str(e)})
        _secret_path = get_secret_path()
        secret_path = f'{settings.MEDIA_ROOT}/{_secret_path}'
        try:
            archive.extractall(secret_path)
        except (rarfile.BadRarFile, zipfile.BadZipFile) as e:
            raise ValidationError({'zipfile': str(e)})

        images = []

        try:
            html_file_path, _root_folder = _validate_and_find_html(secret_path, '', '', '')

            if not html_file_path:
                raise ValidationError({'zipfile': 'Архив должен содержать html файл'})

            _parse_dir(secret_path, images, '', _root_folder.split('/'))

            self._html_file = html_file_path
            self._images = images
            self._secret_path = _secret_path
        except ValidationError as e:
            shutil.rmtree('/'.join(secret_path.split('/')[:-2]), ignore_errors=True)
            raise ValidationError(e)

    def clean(self):
        super().clean()
        if self.html_from_type == self.HTMLFormType.CKEDITOR:
            if not self.html:
                raise ValidationError({'html': 'Это поле не может быть пустым'})
            self.zipfile = None
            return
        if not self.zipfile:
            raise ValidationError({'zipfile': 'Это поле не может быть пустым'})
        prev_instance = self.__class__.objects.get(pk=self.pk) if self.pk else None
        if not self.pk or prev_instance.zipfile != self.zipfile:
            self._parse_and_validate_zipfile()

    class Meta:
        verbose_name = 'Шаблон'
        verbose_name_plural = 'Шаблоны'
